package wsmg;

import edu.indiana.extreme.q.QueueFactory;
import java.net.URI;
import java.sql.SQLException;
import org.activemq.broker.BrokerFactory;
import org.activemq.broker.BrokerService;
import org.xmlpull.v1.builder.XmlElement;
import wsmg.config.WSMGParameter;
import wsmg.db.SubscriptionDB;
import wsmg.measure.profile.RunTimeStatistics;
import wsmg.processors.NotificationConsumer;
import wsmg.processors.PublisherRegistrationManager;
import wsmg.processors.SubscriptionManager;
import wsmg.processors.WSRL_Processor;
import wsmg.processors.WSRP_Processor;
import wsmg.processors.widget.Widget;
import wsmg.processors.widget.WidgetState;
import wsmg.wsrm.axis2.DedicatedSimpleAxis2Server;
import wsmg.wsrm.axis2.DedicatedSimpleSandeshaServer;
import wsmg.wsrm.axis2.SandeshaConfiguration;
import wsmg.xsul.XSULConfiguration;
import wsmg.xsul.XSULDedicatedBrockerService;
import xsul.MLogger;
import xsul.processor.DynamicInfosetProcessorException;
import xsul.soap.SoapUtil;
import xsul.soap11_util.Soap11Util;
import xsul.soap12_util.Soap12Util;
import xsul.util.FastUUIDGen;
import xsul.ws_addressing.WsaMessageInformationHeaders;

/**
 * This class is intended to be independent of the SOAP transport used. Both the
 * XSUL and Axis2 is suppose to use this class as the entry point to the
 * brocker.
 *
 * @author Chathura Herath (cherath@cs.indiana.edu)
 */
public class WSMGService {

    private final static MLogger logger = MLogger.getLogger();

    private WsmgAdapter wsntAdapter;

    private WsmgConfig wsmgConfig = null;

    private SubscriptionManager subscriptionManager = null;

    private NotificationConsumer notificationConsumer = null;

    private PublisherRegistrationManager publisherRegistrationManager = null;

    private WSRP_Processor wsrp_Processor = null;

    private WSRL_Processor wsrl_Processor = null;

    private Widget widget = null;

    private XSULDedicatedBrockerService xsulServer;

    private DedicatedSimpleAxis2Server axis2Server;

    private DedicatedSimpleSandeshaServer sandeshaServer;

//    private SubscriptionDB subscriptionDB = new SubscriptionDB();

    private WSMGService() {

        if (WSMGParameter.queueType == WSMGParameter.EXTREME_QUEUE) {
            try {
                WSMGParameter.OUT_GOING_QUEUE = QueueFactory.createQueue();
            } catch (SQLException e) {
                throw new NotificationException(
                        "Unable to Create Extreme Queue", e);
            }
        }
        System.out.println("Get init()");
        if (WSMGParameter.queueType == WSMGParameter.EXTREME_DISTRIBUTED_QUEUE) {
            try {
                WSMGParameter.OUT_GOING_QUEUE = QueueFactory
                        .createDistrebutedQueue();
                if (WSMGParameter.cleanQueueonStartUp) {
                    System.out.println("Calling cleanup()");
                    WSMGParameter.OUT_GOING_QUEUE.cleanup();
                    System.out.println("Finished Calling cleanup()");
                }
            } catch (SQLException e) {
                throw new NotificationException(
                        "Unable to Create Extreme Queue", e);
            }
        }
        if (WSMGParameter.useEmbeddedActiveMQ) {
            BrokerService broker;
            try {
                broker = BrokerFactory.createBroker(new URI(
                        "xbean:file:../config/activemq.xml"));

                // broker.setDataDirectory(new File(new
                // File("varyi"),"localhost"));
                // configure the broker
                // broker.addConnector("vm://localhost:61616");
                // broker.setDeleteAllMessagesOnStartup(true);

                // broker.start();
                System.out.println("Embedded AMQ started.");
            } catch (Exception e) {
                throw new WsmgRuntimeException(
                        "Failed to start the embedded ActiveMQ broker.", e);
            }

        }
        RunTimeStatistics.setStartUpTime();

    }

    // public WSMGService(URI serverLocation) {
    // this();
    // this.serverLocation = serverLocation;
    // }

    public WSMGService(WsmgConfig wsmgConfig) {

        this();
        this.wsmgConfig = wsmgConfig;
        if (WsmgConstants.BROKER_WEB_SERVICE_STACK_STANDALONE_XSUL == wsmgConfig
                .getBrokerWebServiceStack()) {
            XSULConfiguration xsulConfig = new XSULConfiguration(
                    new SoapUtil[] { Soap12Util.getInstance(),
                            Soap11Util.getInstance() }, wsmgConfig.getPort());
            xsulConfig.setBroker(this);
            this.xsulServer = new XSULDedicatedBrockerService(xsulConfig);
        }

        // this.axis2Server = new DedicatedSimpleAxis2Server(this);

        if (wsmgConfig.isWsrmEnabled()) {
            SandeshaConfiguration sandeshaConfiguration = new SandeshaConfiguration(
                    wsmgConfig.getRmport(), this);
            this.sandeshaServer = new DedicatedSimpleSandeshaServer(
                    sandeshaConfiguration);
        }

    }

    public void start() {
        if (wsmgConfig.getBrokerWebServiceStack() == WsmgConstants.BROKER_WEB_SERVICE_STACK_STANDALONE_XSUL) {
            this.xsulServer.start();
            wsmgConfig.setServiceLocation(xsulServer.getServiceLocation());

        }

        if (wsmgConfig.isWsrmEnabled()) {
            this.sandeshaServer.start();
            this.sandeshaServer.deployServersideServices();
        }

        if (wsmgConfig.getServiceLocation() == null)
            throw new WsmgRuntimeException(
                    "Unable to start broker. Service Location is not resolved properly");

        subscriptionManager = new SubscriptionManager(wsmgConfig);
        //Decide here to make the server start up even there are malformed subscriptions
        try{
        subscriptionManager.checkSubscriptionDB();
        }catch(SoapClientFault e){
        	System.out.println("*****Subscription database has malformed subscriptions. Ignoring them.");
        	e.printStackTrace();
        }
        notificationConsumer = new NotificationConsumer(wsmgConfig,
                subscriptionManager);
        publisherRegistrationManager = new PublisherRegistrationManager(
                wsmgConfig.getServiceLocation());
        wsrp_Processor = new WSRP_Processor();
        wsrl_Processor = new WSRL_Processor();
        widget = new Widget(wsmgConfig.getServiceLocation(),
                subscriptionManager);


    }

    public void shutdown() {
        this.xsulServer.shutdown();

        if (wsmgConfig.isWsrmEnabled()) {
            this.sandeshaServer.shutDown();
           
        }
    }

    public void executeSoapEnvelope(ProcessingContext ctx) throws SoapFault,
            DynamicInfosetProcessorException {
        String messageName = ctx.getSoapBody().getName();
        
 
        // TODO: URGENT replace with reflection based dispatch
        if (WsmgConstants.WIDGET_NS.equals(ctx.getSoapBody().getNamespace())
                || "http://widgets.com/WSRFWidgetService".equals(ctx
                        .getSoapBody().getNamespace().getNamespaceName()) // special
        // hack
        // for
        // IBM client
        ) {
            ctx.setRespMessage(WsmgConstants.builder.newFragment(
                    WsmgConstants.WIDGET_NS, messageName + "Response"));
            if ("createWidget".equals(messageName)) {
                widget.createWidget(ctx.getRespMessage());
            } else if ("generateNotification".equals(messageName)) {
                // send notificaiton to all subscribers
                widget.generateNotification(ctx);
                // ctx.respMessage = null;
            } else if ("getAllSubscriptions".equals(messageName)) {
                subscriptionManager.getAllSubscriptions(ctx);
            } else if ("addNewBroker".equals(messageName)) {
                // subscriptionManager.getAllSubscriptions(ctx);
            } else if ("getAllSubscriptions".equals(messageName)) {
                // subscriptionManager.getAllSubscriptions(ctx);
            } else {
                throw new SoapClientFault("unrecognized widget message "
                        + messageName);
            }
        } else if (WsmgConstants.WSRL_NS.equals(ctx.getSoapBody()
                .getNamespace())) {
            if (ctx.getResourceId() == null && ctx.getSubId() == null) {
                throw new SoapClientFault("missing resource id in "
                        + messageName);
            }
            assert ctx.getResourceState() != null
                    || ctx.getSubscription() != null;
            ctx.setRespMessage(WsmgConstants.builder.newFragment(
                    WsmgConstants.WSRL_NS, messageName + "Response"));
            ctx.getRespMessage().declareNamespace(WsmgConstants.WSRL_NS);
            if ("Destroy".equals(messageName)) {
                if (ctx.getResourceId() != null) {
                    widget.getWidgets().remove(ctx.resourceId);
                } else {
                    SubscriptionState subscription = (SubscriptionState) subscriptionManager
                            .getSubscriptions().get(ctx.subId);
                    WsmgAdapterConnection wsntAdapterConnection = subscription
                            .getWsntAdapterConnection();
                    if (wsntAdapterConnection != null) {
                        wsntAdapterConnection.stop();
                    }
//                    subscriptionManager.getSubscriptions().remove(ctx.subId);
//                    subscriptionDB.delete(ctx.subId);
                    subscriptionManager.unsubscribe(ctx);
                }
            } else if ("SetTerminationTime".equals(messageName)) {
                wsrl_Processor.setTerminationTime(ctx);
            } else {
                throw new SoapClientFault("unsupported WSRL message "
                        + messageName);
            }
        } else if (WsmgConstants.WSRP_NS.equals(ctx.soapBody.getNamespace())) { // TODO:
            // move
            // to
            // subclass
            // ...
            ctx.respMessage = WsmgConstants.builder.newFragment(
                    WsmgConstants.WSRP_NS, messageName + "Response");
            ctx.respMessage.declareNamespace(WsmgConstants.WSRL_NS);
            ctx.respMessage.declareNamespace(WsmgConstants.WSRP_NS);
            if ("GetResourceProperty".equals(messageName)) {
                wsrp_Processor.getResourceProperty(ctx);
            } else if ("GetMultipleResourceProperties".equals(messageName)) {
                wsrp_Processor.getMultipleResourceProperties(ctx);
            } else if ("SetResourceProperties".equals(messageName)) {
                // ok lets see what is sub-operation
                wsrp_Processor.setResourceProperties(ctx);
            } else if ("QueryResourceProperties".equals(messageName)) {
                wsrp_Processor.queryResourceProperties(ctx);
            } else {
                throw new SoapClientFault("unsupported WSRP message "
                        + messageName);
            }
        } else if (WsmgConstants.WSNT_NS.equals(ctx.soapBody.getNamespace())) { // TODO:
            // move
            // to
            // subclass
            // ...
            ctx.respMessage = WsmgConstants.builder.newFragment(
                    WsmgConstants.WSNT_NS, messageName + "Response");
            ctx.respMessage.declareNamespace(WsmgConstants.WSNT_NS);
            if ("Subscribe".equals(messageName)) {
                subscriptionManager.subscribe(ctx);
            } else if ("GetCurrentMessage".equals(messageName)) {
                widget.getCurrentMessage(ctx);
            } else if ("PauseSubcription".equals(messageName)
                    || "PauseSubscription".equals(messageName)) {
                subscriptionManager.pauseSubscription(ctx);
            } else if ("ResumeSubscription".equals(messageName)) {
                subscriptionManager.resumeSubscription(ctx);
            } else if ("Notify".equals(messageName)) {
                notificationConsumer.notify(ctx);
            } else {
                throw new SoapClientFault("unsupported WSBN message "
                        + messageName);
            }
        } else if (WsmgConstants.WSBR_NS.equals(ctx.soapBody.getNamespace())) { // TODO:
            // move
            // to
            // subclass
            // ...
            ctx.respMessage = WsmgConstants.builder.newFragment(
                    WsmgConstants.WSBR_NS, messageName + "Response");
            ctx.respMessage.declareNamespace(WsmgConstants.WSNT_NS);
            ctx.respMessage.declareNamespace(WsmgConstants.WSBR_NS);
            if ("RegisterPublisher".equals(messageName)) {
                publisherRegistrationManager.registerPublisher(ctx);
            } else {
                throw new SoapClientFault("unsupported WSBN message "
                        + messageName);
            }
        } else if (WsmgConstants.WSE_NS.equals(ctx.soapBody.getNamespace())) { // TODO:
            // move
            // to
            // subclass
            // ...
            ctx.respMessage = WsmgConstants.builder.newFragment(
                    WsmgConstants.WSE_NS, messageName + "Response");
            ctx.respMessage.declareNamespace(WsmgConstants.WSE_NS);
            if ("Subscribe".equals(messageName)) {
                subscriptionManager.subscribe(ctx);
            } else if ("Renew".equals(messageName)) {
                // getCurrentMessage(ctx);
            } else if ("GetStatus".equals(messageName)) {
                // getCurrentMessage(ctx);
            } else if ("Unsubscribe".equals(messageName)) {
                subscriptionManager.unsubscribe(ctx);
                /*
                 * SubscriptionState subscription = (SubscriptionState)
                 * subscriptions. get(ctx.subId); WsntAdapterConnection
                 * wsntAdapterConnection = subscription.
                 * getWsntAdapterConnection(); if (wsntAdapterConnection !=
                 * null) { wsntAdapterConnection.stop(); }
                 * subscriptions.remove(ctx.subId);
                 * subscriptionDB.delete(ctx.subId);
                 */
            } else if ("Renew".equals(messageName)) {
                // getCurrentMessage(ctx);
            } else {
                throw new SoapClientFault("unsupported WSE message "
                        + messageName);
            }
        } else { // WSE notification message. Any message MAY be a
            // notification.
                if(WSMGParameter.measureInternalProcessingTime){
            logger.info("*******Got notification time="
                    + System.currentTimeMillis());
                }
            // ctx.message.addAttribute("MessageSequenceNumber","YiHuang's
            // Message");
//            logger.info("########Sending="
//                    + WsmgConstants.builder.serializeToString(ctx.soapBody));
            ctx.respMessage = WsmgConstants.builder.newFragment(
                    WsmgConstants.WSE_NS, messageName + "Response");
            ctx.respMessage.declareNamespace(WsmgConstants.WSE_NS);

            // while(requestXmlMsgCount.getCounterValue()<400000){

            notificationConsumer.notify(ctx);

            // requestXmlMsgCount.addCounter();
            // try {
            // Thread.sleep(50);
            // } catch (InterruptedException e) {
            // // TODO Auto-generated catch block
            // e.printStackTrace();
            // }
            // }//added for stress test
            // --while(requestXmlMsgCount.getCounterValue()<400000){
            if(WSMGParameter.measureInternalProcessingTime){
            logger.info("*******Finished notification time="
                    + System.currentTimeMillis());
            }
            // throw new SoapClientFault("unrecongized message " + messageName +
            // " in namespace " +
            // ctx.message.getNamespace().getNamespaceName());
        }
    }

    public void processHeaders(ProcessingContext ctx) {
        XmlElement envelope = ctx.getEnvelope();
        WsaMessageInformationHeaders wsaHeaders = new WsaMessageInformationHeaders(
                envelope);
        // process resourceId
        XmlElement resourceIdElement = wsaHeaders.getHeaderElement(
                WsmgConstants.WIDGET_NS, WsmgConstants.RESOURCE_ID);

        if (resourceIdElement != null) {
            ctx.setResourceId(resourceIdElement.requiredTextContent());
            ctx.setResourceState((WidgetState) widget.getWidgets().get(
                    ctx.getResourceId()));
            if (ctx.getResourceState() == null) {
                throw new SoapClientFault(
                        "no resource associated with recource id "
                                + ctx.getResourceId());
            }
        }

        // prcess action
        XmlElement subIdElement = null;

        XmlElement actionEl = wsaHeaders.getHeaderElement(null, WsmgConstants.WSA_ACTION_HEADER_NAME);
        String action = null;
        if (actionEl != null) {
            action = actionEl.requiredTextContent();
        } else {
            if(ctx.getTopicFromUrl() != null) { // allow no SOAP WSA Action header if topic was in URL
                logger.info("adding WSA Action and MessageId for URL topic "+ctx.getTopicFromUrl());
                wsaHeaders.setAction(URI.create("http://www.extreme.indiana.edu/WseNotificationFromUrl/"+ctx.getTopicFromUrl()));
                wsaHeaders.setMessageId(URI.create("uuid:"+FastUUIDGen.nextUUID()));
                return;
            } else {
                throw new SoapClientFault("Action required in soap header.");
            }
        }
        if (action.startsWith(WsmgConstants.WSN1_0_ActionPrefix)) {
            // check WSE or WSNT by action

            subIdElement = wsaHeaders.getHeaderElement(WsmgConstants.WSE_NS,
                    "Identifier");
        } else if (action.startsWith("http://schemas.xmlsoap.org/ws")) { // WSE
            subIdElement = wsaHeaders.getHeaderElement(WsmgConstants.WSE_NS,
                    "Identifier");
        }

        if (subIdElement != null) {
            ctx.setSubId(subIdElement.requiredTextContent());
            ctx.setSubscription(((SubscriptionState) subscriptionManager
                    .getSubscriptions().get(ctx.getSubId())));
            // TODO: should send error message to the client instead of just
            // throwing exception
            if (ctx.getSubscription() == null) {
                throw new SoapClientFault("no subscription associated with id "
                        + ctx.getSubId());
            }
        }
    }

    public String getServer() {
        return this.xsulServer.getServer().getLocation();
    }

    public XSULDedicatedBrockerService getXsulServer() {
        return xsulServer;
    }

}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 *
 * Copyright (C) 2004 The Trustees of Indiana University. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 *
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 *
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 *
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 *
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 *
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 *
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 *
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */

