package wsmg.util;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Iterator;
import javax.xml.namespace.QName;
import org.xmlpull.v1.XmlSerializer;
import org.xmlpull.v1.builder.XmlAttribute;
import org.xmlpull.v1.builder.XmlElement;
import org.xmlpull.v1.builder.XmlInfosetBuilder;
import org.xmlpull.v1.builder.XmlNamespace;
import wsmg.ProcessingContext;
import wsmg.SoapClientFault;
import wsmg.WsmgConstants;
import xsul.MLogger;
import xsul.XmlConstants;
import xsul.XsulException;
import xsul.util.XsulUtil;

public class WsmgUtil {
    private static final String BODY = "Body";
    
    private static final String ACTION = "Action";
    
    private static final String MESSAGE_ID = "MessageID";
    
    private static final String XMLNS_WIDGET = "xmlns:widget";
    
    private static final String DIALECT = "Dialect";
    
    private static final String TOPIC = "Topic";
    
    private static final String TO = "To";
    
    private static final String HEADER = "Header";
    
    private static final String ENVELOPE = "Envelope";
    
    private static final String WSNT = "wsnt";
    
    private static final String XSI = "xsi";
    
    private static final String WA48 = "wa48";
    
    private static final String S = "S";
    
    private static final String HTTP_WIDGETS_COM = "http://widgets.com";
    
    private static final String HTTP_WWW_IBM_COM_XMLNS_STDWIP_WEB_SERVICES_WS_TOPICS_TOPIC_EXPRESSION_SIMPLE = "http://www.ibm.com/xmlns/stdwip/web-services/WS-Topics/TopicExpression/simple";
    
    private static final String HTTP_WWW_IBM_COM_XMLNS_STDWIP_WEB_SERVICES_WS_BASE_NOTIFICATION = "http://www.ibm.com/xmlns/stdwip/web-services/WS-BaseNotification";
    
    private static final String HTTP_WWW_W3_ORG_2001_XMLSCHEMA_INSTANCE = "http://www.w3.org/2001/XMLSchema-instance";
    
    private static final String HTTP_SCHEMAS_XMLSOAP_ORG_WS_2004_08_ADDRESSING = "http://schemas.xmlsoap.org/ws/2004/08/addressing";
    
    private static final String HTTP_SCHEMAS_XMLSOAP_ORG_SOAP_ENVELOPE = "http://schemas.xmlsoap.org/soap/envelope/";
    
    private final static XmlInfosetBuilder builder = XmlConstants.BUILDER;
    
    private final static MLogger logger = MLogger.getLogger();
    
    public final static XmlNamespace EXTREME_NS = builder.newNamespace(
        "extreme", "http://extreme.indiana.edu");
    
    /**
     * if pure String, if WSNT and UseNotify: ==><wsnt:message>test message</wsnt:Message>
     * if WSE or UseNotify=false ==><widget:TestNotification>test message</widget:TestNotification>
     * if String is an Xml (starts with <) if WSNT and UseNotify: =><wsnt:message>String
     * message</wsnt:Message> if WSE or UseNotify=false =>keep the Xml.
     */
    // Updated: 1/1/2006 xsd:any cannot be a pure String <wsnt:Messenge>My
    // String</wsnt:Messenger> doesn't follow specification
    public static XmlElement convertStringMessageToXmlElement(String message) {
        XmlElement messageToNotify = null;
        if (message.trim().startsWith("<")) { // Message is an XML Element
            XmlElement messageEl = builder
                .parseFragmentFromReader(new StringReader(message));
            
            messageToNotify = messageEl;
            
            return messageToNotify;
            
        } else {
            messageToNotify = builder.newFragment(EXTREME_NS,
                                                  "StringNotification");
            
            messageToNotify.addChild(message);
        }
        return messageToNotify;
    }
    
    public static String formatURLString(String url) {
        if (url.indexOf("//") < 0) {
            url = "http://" + url; // use default http
        }
        return url;
    }
    
    public static boolean sameStringValue(String stringA, String stringB){
        if(stringA==null){
            if(stringB==null){
                return true;
            }else{
                return false;
            }
        }
        //StringA!=null
        if(stringB==null) return false;
        if(stringA.compareTo(stringB)==0) {
            return true;
        }else{
            return false;
        }
    }
    
    public static String getSoapBodyContent(String wholeMessage) {
        // parse the received String message and create XmlElement object
        XmlElement messageEl = builder
            .parseFragmentFromReader(new StringReader(wholeMessage));
        
        XmlElement messageContentEl = messageEl.requiredElement(
            WsmgConstants.SOAP_NS, BODY);
        
        if (logger.isFinestEnabled()) {
            logger.finest("Message="
                              + builder.serializeToString(messageContentEl));
        }
        // messageContentEl is the XML element containing the SOAP body tag.
        // The following codes get the content of the SOAP body.
        String messageContentAsXml = getContentInsideAnXmlElementAsString(messageContentEl);
        
        return messageContentAsXml;
    }
    
    public static String getContentInsideAnXmlElementAsString(
        XmlElement messageContentEl) {
        StringWriter sw = null;
        try {
            XmlSerializer xs = builder.getFactory().newSerializer();
            sw = new StringWriter();
            xs.setOutput(sw);
            for (Iterator i = messageContentEl.children(); i.hasNext();) {
                builder.serialize(i.next(), xs);
            }
            xs.flush();
        } catch (Exception ex2) {
            throw new XsulException("could not convert message content to XML",
                                    ex2);
        }
        
        String messageContentAsXml = sw.toString();
        return messageContentAsXml;
    }
    
    public static String getHostIP() {
        InetAddress localAddress = null;
        try {
            localAddress = InetAddress.getLocalHost();
        } catch (UnknownHostException ex) {
            System.out.println("Error - unable to resolve localhost");
        }
        // Use IP address since DNS entry cannot update the laptop's entry
        // promptly
        String hostIP = localAddress.getHostAddress();
        return hostIP;
    }
    
    /**
     *
     * @return localString
     */
    
    public static String getTopicLocalString(XmlElement topicEl) {
        XmlAttribute dialectAttribute = topicEl.attribute("dialect");
        if (dialectAttribute == null) {
            dialectAttribute = topicEl.attribute(DIALECT); //try upper case
        }
        if (dialectAttribute == null) {
            throw new SoapClientFault("dialect is required for subscribe");
        }
        String dialectString = dialectAttribute.getValue();
        
        if (!dialectString
                
                .equals(WsmgConstants.TOPIC_EXPRESSION_SIMPLE_DIALECT)) {
            //    System.out.println("***Unkown dialect: " + dialectString);
            throw new SoapClientFault("Unkown dialect: " + dialectString);
            
        }
        String topicExpression = topicEl.requiredTextContent();
        QName topicExpressionQName = XsulUtil.toQName(topicEl, topicExpression);
        String topicLocalString = topicExpressionQName.getLocalPart();
        System.out.println("******Topic="+topicLocalString);
        return topicLocalString;
    }
    
    /**
     *
     * @return localString
     */
    public static String getXPathString(XmlElement xpathEl) {
        XmlAttribute dialectAttribute = xpathEl.attribute("dialect");
        if (dialectAttribute == null) {
            dialectAttribute = xpathEl.attribute(DIALECT); //try upper case
        }
        if (dialectAttribute == null) {
            throw new SoapClientFault("dialect is required for subscribe");
        }
        String dialectString = dialectAttribute.getValue();
        if (!dialectString
                .equals(WsmgConstants.XPATH_DIALECT)) {
            //  System.out.println("***Unkown dialect: " + dialectString);
            throw new SoapClientFault("Unkown dialect: " + dialectString);
        }
        String xpathLocalString = xpathEl.requiredTextContent();
        return xpathLocalString;
    }
    
    
    
    // From http://www.rgagnon.com/javadetails/java-0052.html
    public static String readFile(String filename) throws IOException {
        String lineSep = System.getProperty("line.separator");
        BufferedReader br = new BufferedReader(new FileReader(filename));
        String nextLine = "";
        StringBuffer sb = new StringBuffer();
        while ((nextLine = br.readLine()) != null) {
            sb.append(nextLine);
            //
            // note:
            // BufferedReader strips the EOL character.
            //
            sb.append(lineSep);
        }
        return sb.toString();
    }
    
    public static XmlElement createSoapEnvelop(XmlElement bodyContentXML, String wsaTO,
                                               String topicElement, String messageID, String action) {
        XmlNamespace soapEnvNameSpace = builder.newNamespace(S,
                                                             HTTP_SCHEMAS_XMLSOAP_ORG_SOAP_ENVELOPE);
        XmlNamespace wsaNameSpace = builder.newNamespace(WA48,
                                                         HTTP_SCHEMAS_XMLSOAP_ORG_WS_2004_08_ADDRESSING);
        XmlNamespace xsi = builder.newNamespace(XSI,
                                                HTTP_WWW_W3_ORG_2001_XMLSCHEMA_INSTANCE);
        
        XmlNamespace wsntNameSpace = builder
            .newNamespace(WSNT,
                          HTTP_WWW_IBM_COM_XMLNS_STDWIP_WEB_SERVICES_WS_BASE_NOTIFICATION);
        
        XmlElement envelop = builder.newFragment(soapEnvNameSpace, ENVELOPE);
        
        XmlElement header = envelop.addElement(soapEnvNameSpace, HEADER);
        if (wsaTO != null) {
            XmlElement wsaTOElement = header.addElement(wsaNameSpace, TO);
            wsaTOElement.addChild(wsaTO);
        }
        if (topicElement != null) {
            XmlElement topic = header.addElement(builder.parseFragmentFromReader(new StringReader(topicElement)));
            //            topic
            //                    .addAttribute(DIALECT,
            //                            HTTP_WWW_IBM_COM_XMLNS_STDWIP_WEB_SERVICES_WS_TOPICS_TOPIC_EXPRESSION_SIMPLE);
            //            topic.addAttribute(XMLNS_WIDGET, HTTP_WIDGETS_COM);
            //            topic.addChild(topicString);
        }
        if (messageID != null) {
            XmlElement messageIDElement = header.addElement(wsaNameSpace,
                                                            MESSAGE_ID);
            messageIDElement.addChild(messageID);
        }
        if (action != null) {
            XmlElement actionElement = header
                .addElement(wsaNameSpace, ACTION);
            actionElement.addChild(action);
        }
        
        XmlElement body = envelop.addElement(soapEnvNameSpace, BODY);
        body.addChild(bodyContentXML);
        
        return envelop;
    }
    
    public static void setTopicFromRequestPath(ProcessingContext ctx, String topicPath) {
        if(topicPath == null) return;
        if(topicPath.length() == 0) return;
        if(topicPath.startsWith("/")) {
            topicPath = topicPath.substring(1);
            if(topicPath.length() == 0) return;
        }
        if( topicPath.startsWith(WsmgConstants.TOPIC_PREFIX) ) {
            topicPath = topicPath.substring(WsmgConstants.TOPIC_PREFIX.length());
        }
        ctx.setTopicFromUrl(topicPath);
        logger.info("topicFromUrl="+topicPath);
    }
    
   
}
