/* -*- mode: Java; c-basic-offset: 4; indent-tabs-mode: nil; -*-  //------100-columns-wide------>|*/
/*
 * Copyright (c) 2002-2004 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the licence.
 *
 * $Id: BaseClient.java,v 1.2 2006/01/04 17:47:25 cherath Exp $
 */

package wsmg.xdir;

import java.security.cert.X509Certificate;

import org.globus.gsi.GlobusCredential;
import org.globus.gsi.GlobusCredentialException;
import org.globus.gsi.TrustedCertificates;

import soaprmi.RemoteException;
import xdrs.XDirectoryService;
import xdrs_impl.Run;
import xdrs_util.XmlNodeUtil;
import xsul.MLogger;
import xsul.XsulException;

/**
 * Base class to hold common functonality for client side access to XDRS.
 * 
 * @version $Revision: 1.2 $ $Date: 2006/01/04 17:47:25 $ (GMT)
 * @author <a href="http://www.extreme.indiana.edu/~aslom/">Aleksander Slominski</a>
 * @author Gopi
 */
public abstract class BaseClient {
	private final static MLogger logger = MLogger.getLogger();

	boolean QUIET = false;

	XDirectoryService directoryService;

	boolean secure = true;

	boolean xdrs_internal_test = System.getProperty("xdrs_internal_test") != null;

	String key = System.getProperty(Run.XDRS_KEY);

	String cert = System.getProperty(Run.XDRS_CERT);

	String location;

	String internalServerPort;

	/* provide prefix to print with all messages used ot identify client */
	public abstract String getPrefix(); // { return "XDRS BaseClient: "; }

	/* do actual client work here */
	public abstract void doWork() throws Exception;

	public void run(String[] args) {
		try {
			msg("starting ...");
			processArgs(args);
			msg("will try to connect to " + location + " with "
					+ (secure ? "" : "no ") + "security");

			directoryService = getXdrsStartpoint();

			doWork();

		} catch (Exception ex) {
			ex.printStackTrace();
			System.err.println(getPrefix() + "client execution failed");
			System.exit(1);
		}
	}

	public void msg(String msg) {
		if (!QUIET) {
			System.err.println(getPrefix() + msg);
		}
	}

	public void usage(String msg) {
		if (msg != null) {
			System.err.println("ERROR: " + msg + "\n");
		}
		System.out.println(getUsage());
		System.exit(2);

	}

	public String getUsage() {
		final String USAGE = "usage: java "
				+ getClass().getName()
				+ " http://service.location OPTIONS \n"
				+
				// "[--context bind_context] \n" +
				"[--cert certificate_corresponsding_to_key "
				+ "--key unencrypted_private_key]\n"
				+ "[--port NNNN] \t port on which to start INTERNAL service\n"
				+ "[--quiet] \t do not print anything unless it is error\n"
				+ "[--unsecure] \t do not use security when accessing service\n"
				+
				// "[--rebind name] \t name to which bind service\n"+
				"";
		return USAGE;
	}

	public int unknownArg(String[] args, int argsCount) throws Exception {
		msg("skipping unknown argument " + args[argsCount]);
		return argsCount;
	}

	public void processArgs(String[] args) throws Exception {
		// boolean xdrs_internal_test = false;

		int argsCount = 0;
		while (argsCount < args.length) {
			String arg = args[argsCount];
			if (arg.startsWith("http://")) {
				location = arg;
			} else if (args[argsCount].equals("--quiet")) {
				QUIET = true;
			} else if (args[argsCount].equals("--unsecure")) {
				secure = false;
			} else if (args[argsCount].equals("--key")) {
				key = args[++argsCount];
			} else if (args[argsCount].equals("--cert")) {
				cert = args[++argsCount];
			} else if (arg.equals("--port")) {
				internalServerPort = args[++argsCount];
			} else {
				argsCount = unknownArg(args, argsCount);
			}
			++argsCount;
		}

		if (xdrs_internal_test) {
			if (internalServerPort != null) {
				msg("launching in-process internal XDRS on port "
						+ internalServerPort);
				if (secure) {
					xdrs_impl.Run.main(new String[] { "--port",
							internalServerPort });
				} else {
					xdrs_impl.Run.main(new String[] { "--port",
							internalServerPort, "--unsecure" });
				}
				msg("warning: when running internal tests TCP port is automatically set!");
				internalServerPort = "" + xdrs_impl.Run.getServerPort();
			} else {
				msg("launching in-process internal XDRS");
				if (secure) {
					Run.main(new String[] {}); // {"--port","33489"});
				} else {
					Run.main(new String[] { "--unsecure" });
				}
				internalServerPort = "" + xdrs_impl.Run.getServerPort();
			}
		} else {
			if (location == null) {
				throw new RuntimeException(
						"URL with service location is required");
			}
		}
		if (key == null) {
			key = System.getProperty(Run.XDRS_KEY);
		}
		if (cert == null) {
			cert = System.getProperty(Run.XDRS_CERT);
		}

		if (key != null || cert != null) {
			if (key == null || cert == null) {
				usage(getPrefix() + " both certificate and key must be set");
			}
			msg("will use key=" + key + " and cert=" + cert
					+ " instead of default proxy");
		}

		assert internalServerPort != null || location != null;
		if (location == null && internalServerPort != null) {
			location = "http://localhost:" + internalServerPort;
		}
	}

	/**
	 * Create startpoint to XDRS that is configured according to command line
	 * options and system properties and print what is security context
	 * configuration
	 */
	public XDirectoryService getXdrsStartpoint() throws XsulException,
			GlobusCredentialException, RemoteException {
		XDirectoryService directoryService;
		if (!secure) {
			// get access to directory service located at URL passed from
			// command line
			directoryService = XmlNodeUtil.getDirectoryServiceProxy(location);
		} else {

			GlobusCredential cred;
			if (key != null || cert != null) {
				msg("using certificate " + cert + " and key " + key);
				cred = new GlobusCredential(cert, key);
			} else {
				msg("using default CoG credential");
				cred = GlobusCredential.getDefaultCredential();
			}

			// msg("loading default globus credential (grid proxy) ...");
			// GlobusCredential cred = GlobusCredential.getDefaultCredential();
			logger.config("cred=" + cred);
			msg("using globus credential with DN=" + cred.getSubject());

			TrustedCertificates tc = TrustedCertificates
					.getDefaultTrustedCertificates();
			if (tc == null) {
				throw new XsulException(
						getPrefix()
								+ " can not be started as trusted certificates were not found");
			}

			X509Certificate[] trustedCerts = tc.getCertificates();
			msg("list of trusted of certifcates");
			for (int i = 0; i < trustedCerts.length; i++) {
				msg("trusted cert " + i + "=" + trustedCerts[i].getSubjectDN());
			}

			directoryService = XmlNodeUtil.getDirectoryServiceProxy(location,
					cred, trustedCerts);
			msg("using directory service " + location + " with credential "
					+ cred.getIssuer());
		}
		return directoryService;
	}

}
/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2002-2004 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */
